from .shift_left import shift_left

def add_input_pos_to_func_str(func_str:str, submodule_name:str, new_arg_string:str):
    lines = func_str.split('\n')
    found_submodule_forward = False
    submodule_indent = -1

    # Find the forward pass of the specified submodule
    for i, line in enumerate(lines):
        if found_submodule_forward:
            submodule_indent = max(submodule_indent, len(line) - len(line.lstrip()))
            if ')' in line:
                break
        elif submodule_name in line and (submodule_name + '(' in line or submodule_name + '.forward(' in line):
            found_submodule_forward = True

    if not found_submodule_forward:
        raise ValueError("Submodule forward pass not found.")

    # Add the new argument to the forward pass
    for i in range(i, len(lines)):
        if lines[i].strip().startswith(')'):
            lines.insert(i, f"{' ' * submodule_indent}{new_arg_string},")
            break

    res_str = '\n'.join(lines)

    return shift_left(res_str)


import types

def add_str_as_func(obj:object, method_name:str, func_str:str, imports:list[str] = []):
    import_str = "\n".join(imports)

    # Combine imports and func string
    complete_func_str = import_str + "\n" + func_str

    func_code = compile(complete_func_str, "<string>", "exec")
    namespace = {}
    exec(func_code, namespace)

    # Extract the func from the namespace
    my_func = namespace[method_name]

    # Attach the func to the object
    setattr(obj, method_name, types.MethodType(my_func, obj))

    return obj


from .shift_left import shift_left

def modify_function_block(method_str:str, line_to_search:str, num_lines_to_replace:int, replace_list:list[str]) -> str:
    # Split the source code of the function
    source_code_split = method_str.split('\n')

    # Find the index of the line to search for
    line_index = None
    for index, line in enumerate(source_code_split):
        if line_to_search in line:
            line_index = index
            break

    if line_index is not None:
        # Count the number of spaces that the line is indented over
        indent = len(source_code_split[line_index]) - len(source_code_split[line_index].lstrip())

        # Adjust each element in replace_list by the calculated indent
        for i in range(len(replace_list)):
            replace_list[i] = ' ' * indent + replace_list[i]

        # Replace lines from line_index to line_index + num_lines_to_replace with replace_list
        source_code_split = source_code_split[:line_index] + replace_list + source_code_split[line_index + num_lines_to_replace:]

    # Join the modified source code and return
    modified_code = '\n'.join(source_code_split)
    return shift_left(modified_code)


from torch import nn
from transformers import AutoModelForCausalLM

def load_int8(model_name:str) -> nn.Module:
    model = AutoModelForCausalLM.from_pretrained(model_name, load_in_8bit=True)
    return model


from typing import Optional, Tuple, Dict, Any
import torch
from transformers.cache_utils import Cache

class KVCache(Cache):
    """
    Static Cache class to be used with `torch.compile(model)`.

    Parameters:
        config (`PretrainedConfig):
            The configuration file defining the `max_position_embeddings`, `hidden_size` and `num_attention_heads`
            required to initialize the static cache.
        max_batch_size (`int`):
            The maximum batch size with which the model will be used.
        max_cache_len (`int`):
            The maximum sequence length with which the model will be used.
        device (`torch.device`):
            The device on which the cache should be initialized. Should be the same as the layer.
        dtype (*optional*, defaults to `torch.float32`):
            The default `dtype` to use when initializing the layer.
    """

    def __init__(self, config, max_batch_size: int, max_cache_len: int, device, dtype=None) -> None:
        super().__init__()
        self.max_batch_size = max_batch_size
        self.max_cache_len = config.max_position_embeddings if max_cache_len is None else max_cache_len
        # Some model define a custom `head_dim` != config.hidden_size // config.num_attention_heads
        self.head_dim = (
            config.head_dim if hasattr(config, "head_dim") else config.hidden_size // config.num_attention_heads
        )

        self.dtype = dtype if dtype is not None else torch.float32
        self.num_key_value_heads = (
            config.num_key_value_heads if hasattr(config, "num_key_value_heads") else config.num_attention_heads
        )

        cache_shape = (max_batch_size, self.num_key_value_heads, self.max_cache_len, self.head_dim)
        self.key_cache: torch.Tensor = torch.zeros(cache_shape, dtype=self.dtype, device=device)
        self.value_cache: torch.Tensor = torch.zeros(cache_shape, dtype=self.dtype, device=device)
        min_value = torch.finfo(self.dtype).min
        self.full_mask = torch.zeros((max_batch_size, self.num_key_value_heads, self.max_cache_len, self.max_cache_len), dtype=self.dtype, device=device)
        min_indices = torch.triu(torch.ones((max_batch_size, self.num_key_value_heads, self.max_cache_len, self.max_cache_len), dtype=torch.bool), diagonal=1)
        self.full_mask[min_indices] = min_value

    def update(
        self,
        key_states: torch.Tensor,
        value_states: torch.Tensor,
        layer_idx: int,
        full: Optional[bool] = False,
        **cache_kwargs: Optional[Dict[str, Any]],
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Updates the cache with the new `key_states` and `value_states` for the layer `layer_idx`.
        It is VERY important to index using a tensor, otherwise you introduce a copy to the device.

        Parameters:
            key_states (`torch.Tensor`):
                The new key states to cache.
            value_states (`torch.Tensor`):
                The new value states to cache.
            layer_idx (`int`):
                The index of the layer to cache the states for. Kept for backward compatibility
            cache_kwargs (`Dict[str, Any]`, `optional`):
                Additional arguments for the cache subclass. The `StaticCache` just needs the `q_len`
                to know how much of the cache it should overwrite.

        Return:
            A tuple containing the updated key and value states.
        """
        new_cache_positions = cache_kwargs.get("input_pos")
        k_out = self.key_cache
        v_out = self.value_cache

        k_out[:, :, new_cache_positions] = key_states
        v_out[:, :, new_cache_positions] = value_states

        self.partial_mask = self.full_mask[:, :, new_cache_positions]
        return k_out, v_out

    def get_seq_length(self, layer_idx: Optional[int] = 0) -> int:
        """Returns the sequence length of the cached states that were seen by the model. `layer_idx` kept for BC"""
        return self.seen_tokens

    def get_usable_length(self, new_sequence_length=None, layer_idx: Optional[int] = 0) -> int:
        return self.seen_tokens

    def get_max_length(self) -> Optional[int]:
        """Returns the maximum sequence length of the cached states. DynamicCache does not have a maximum length."""
        return self.max_cache_len

    def reorder_cache(self, beam_idx: torch.LongTensor):
        """Reorders the cache for beam search, given the selected beam indices."""
        device = self.key_cache.device
        self.key_cache = self.key_cache.index_select(0, beam_idx.to(device))
        device = self.value_cache.device
        self.value_cache = self.value_cache.index_select(0, beam_idx.to(device))

    def to_legacy_cache(self):
        """Dummy func for BC. We have to keep it because otherwise the call in the forward of models will break it"""
        return None


def shift_left(func_str:str, spaces:int=4) -> str:
    # Split the func string into lines
    lines = func_str.split('\n')

    # Shift each line to the left by the specified number of spaces
    for i in range(len(lines)):
        lines[i] = lines[i][spaces:]

    # Join the lines back into a string
    result_str = '\n'.join(lines)

    return result_str


def add_default_parameter(func_str, func_name, default_parameter_name, parameter_type, default_value, has_terminating_comma: bool = False):
    lines = func_str.split('\n')
    signature_end = None
    num_spaces = -1

    # Find the line containing the function name
    for i, line in enumerate(lines):
        if f"def {func_name}" in line:
            start_index = i
            break

    # Proceed with the rest of the function string after finding the function name
    for i in range(start_index, len(lines)):
        num_spaces = max(num_spaces, len(lines[i]) - len(lines[i].lstrip()))
        if ')' in lines[i]:
            signature_end = i
            break

    if signature_end is not None and num_spaces != -1:
        new_param_str = f"{default_parameter_name}: {parameter_type} = {default_value}"
        indent = ' ' * num_spaces
        new_signature = f"{indent}{new_param_str}"
        lines.insert(signature_end, new_signature)

        # Append comma to the line preceding the signature end
        if not has_terminating_comma:
            if signature_end > 0:
                lines[signature_end - 1] += ','

    return '\n'.join(lines[start_index:])


import torch

def timed(fn):
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    start.record()
    result = fn()
    end.record()
    torch.cuda.synchronize()
    return result, start.elapsed_time(end) / 1000


#In order for speculative decoding to work, the vocabulary for two models must be the same, i.e. the dictionary with id keys and string tokens must be the same.
import types
import torch
import torch.nn as nn
from ..KVCache.KVCacheModel import KVCacheModel

#ok, here's the key behind speculative decoding. We have two models, Mq the small model and Mp the large model. 
#1. Run Mq on prefix and obtain the distribution for x1 q(x).
#2. Run Mp on prefix and prefix + x1 concurrently to get the distributions for x2 and p(x).
#3. If x1 is rejected by Mp, reject and resample x1 from an altered distribution, otherwise keep x1 and x2. 

def speculative_decode_eager(
    self,
    cur_tokens:torch.Tensor,
    speculate_k:int,
    **kwargs
) -> torch.Tensor:

    device = cur_tokens.device

    draft_model_sampling_kwargs = kwargs.get("draft_model_decoding_kwargs", {})

    decode_input = cur_tokens

    assert hasattr(self, "draft_model"), "You did not prepare your model properly for speculative decoding. Make sure that you add a draft model."
    draft_model = self.draft_model
    
    draft_tokens, draft_prob = draft_model.decode_function(input_ids=decode_input, length=speculate_k, **draft_model_sampling_kwargs)

    assert len(draft_tokens.shape) == 2 and len(draft_prob.shape) == 2, "Your draft tokens must have shape (1, seq_len) and draft_prob must have shape (seq_len, vocab_size)."

    model_sampling_kwargs = kwargs.get("model_forward_kwargs", {})
    full_tokens = torch.cat([decode_input, draft_tokens], dim=-1).to(device)
    with torch.no_grad():
        model_logits = self.forward(full_tokens, **model_sampling_kwargs).logits
    model_logits = model_logits.squeeze(0)[-draft_tokens.shape[1]-1:, :]
    model_prob = torch.nn.functional.softmax(model_logits, dim=-1)
    
    assert len(model_prob.shape) == 2, "Your model_prob must have shape (seq_len, vocab_size)."

    assert len(model_prob) == len(draft_prob), "In order for speculative decoding to work, the main model must generate the same number of tokens as the draft model."

    p = model_prob[torch.arange(0, speculate_k, device=device), draft_tokens]
    q = draft_prob[torch.arange(0, speculate_k, device=device), draft_tokens]

    ratio = p / q
    rand = torch.rand_like(ratio)

    reject_locations = (rand > ratio).nonzero(as_tuple=True)[0]

    if reject_locations.shape[0] != 0:
        n = reject_locations[0].item()
        p = draft_prob[n]
        q = model_prob[n]
        new = q - p
        new = torch.where(new > 0, new, 0.0)
        last_token = torch.Tensor([self.sample(new)]).to(device)
        
        return torch.cat([draft_tokens[:, :n], last_token.unsqueeze(0)], dim=-1).long()
    else: #we accept all tokens from the draft model
        last_token = torch.Tensor([self.sample(model_prob[-1])]).to(device)
        return torch.cat([draft_tokens, last_token.unsqueeze(0)], dim=-1).long()

def speculative_decode_kv_cache(
    self,
    uncached_tokens:torch.Tensor,
    speculate_k:int,
    **sampling_kwargs
) -> torch.Tensor:

    assert isinstance(self, KVCacheModel), "Your model must be a KVCache model in order to call speculative_decode_kv_cache."

    device = uncached_tokens.device

    assert hasattr(self, "draft_model"), "You did not prepare your model properly for speculative decoding. Make sure that you add a draft model."
    draft_model = self.draft_model
    
    assert isinstance(draft_model, KVCacheModel), "Your draft model muut be a KVCache model in order to call speculative_decode_kv_cache."

    draft_tokens, draft_prob = draft_model.decode_function(uncached_input_ids=uncached_tokens, length=speculate_k, return_text=True)

    assert len(draft_tokens.shape) == 2 and len(draft_prob.shape) == 2, "Your draft tokens must have shape (1, seq_len) and draft_prob must have shape (seq_len, vocab_size)."

    full_tokens = torch.cat([uncached_tokens, draft_tokens], dim=-1).to(device)
    with torch.no_grad():
        self.forward(full_tokens)
    model_logits = self._prob_history[:, -speculate_k-1:, :].squeeze(0)
    model_prob = torch.nn.functional.softmax(model_logits, dim=-1)

    p = model_prob[torch.arange(0, speculate_k, device=device), draft_tokens]
    q = draft_prob[torch.arange(0, speculate_k, device=device), draft_tokens]

    ratio = p / q
    rand = torch.rand_like(ratio)

    reject_locations = (rand > ratio).nonzero(as_tuple=True)[0]

    if reject_locations.shape[0] != 0:
        n = reject_locations[0].item()
        p = draft_prob[n]
        q = model_prob[n]
        new = q - p
        new = torch.where(new > 0, new, 0.0)
        last_token = self.sample(new, **sampling_kwargs).to(device)
        self.rollback_cache(self._cached_len - (full_tokens.shape[-1] - n) + 1)
        draft_model.rollback_cache(draft_model._cached_len - (draft_tokens.shape[-1] - n) + 1)
        return torch.cat([draft_tokens[:, :n], last_token], dim=-1).long() #maybe? This is what my gut says.
    else: #we accept all tokens from the draft model
        last_token = self.sample(model_prob[-1], **sampling_kwargs).to(device)
        #self.rollback_cache(n + len(cur_tokens) + 1)
        draft_model(input_ids=draft_tokens[:, -1].view(1, 1))
        #draft_model.rollback_cache(draft_model._cached_len + 1) #This is a little bit suspect, might have to change this - why are we rejecting the last token when it works?
        #assume that draft_model already has a kv cache attached.
        return torch.cat([draft_tokens, last_token], dim=-1).long() #intuitively, this makes sense.

def generate(self, cur_tokens:torch.Tensor, max_tokens:int, speculate_k:int, **sampling_kwargs) -> torch.Tensor:

    assert len(cur_tokens.shape) == 2 and cur_tokens.shape[0] == 1, "Your batch size must be 1"

    assert hasattr(self, "speculative_decode"), "You must attach speculative decoding as a method of the model"

    while len(cur_tokens[0]) < max_tokens:
        new_tokens = self.speculative_decode(cur_tokens, speculate_k, **sampling_kwargs)
        cur_tokens = torch.cat((cur_tokens, new_tokens), dim=1).to(torch.long)

    return cur_tokens

def generate_kv_cache(self, cur_tokens:torch.Tensor, max_tokens:int, speculate_k:int, **sampling_kwargs) -> torch.Tensor:
    assert isinstance(self, KVCacheModel), "Your model must be a KVCache model for this to work."
    assert hasattr(self, "draft_model") and isinstance(self.draft_model, KVCacheModel), "You must have a draft model and it must be a KVCacheModel for the generate() method."
    assert len(cur_tokens.shape) == 2 and cur_tokens.shape[0] == 1, "Your batch size must be 1"
    assert hasattr(self, "speculative_decode"), "You must attach speculative decoding as a method of the model"

    device = cur_tokens.device

    #prefill phase
    self.draft_model.prefill(cur_tokens)
    new_q = self.prefill(cur_tokens)

    new_token = self.sample(new_q, **sampling_kwargs).to(device)

    cur_tokens = torch.cat((cur_tokens, new_token), dim=-1).to(torch.long)

    while len(cur_tokens[0]) < max_tokens:
        uncached_tokens = cur_tokens[:, -1].view(1, 1)
        new_tokens = self.speculative_decode(uncached_tokens, speculate_k, **sampling_kwargs)
        cur_tokens = torch.cat((cur_tokens, new_tokens), dim=-1).to(torch.long)

    self.clear()
    self.draft_model.clear()

    return cur_tokens    

def add_speculative_decoding(model:nn.Module, draft_model:nn.Module) -> nn.Module:
    model.draft_model = draft_model

    model.speculative_decode = types.MethodType(speculative_decode_kv_cache, model)
    model.generate = types.MethodType(generate_kv_cache, model)
    return model




    
    
    

    







import types
from typing import Callable
import torch
from torch import nn
from transformers.models.bloom.modeling_bloom import BloomForCausalLM
from transformers import AutoModelForCausalLM
from GPTFast.Helpers.Class import *
from GPTFast.Helpers.String import *
from .KVCache import KVCache

INFERENCE_BATCH_SIZE = 1

class KVCacheModel(nn.Module):
    def __init__(self, model:nn.Module, sample_fn:Callable[..., torch.Tensor], max_length:int, cache_config:dict, dtype:torch.dtype):
        super().__init__()
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.sample = types.MethodType(sample_fn, self)

        assert not isinstance(model, BloomForCausalLM), "Bloom models currently have an unsupported kv cache shape."

        self._model = self.add_static_cache_to_model(model, cache_config, max_length, dtype, self.device)
        config = self._model.config
        self._max_length = max_length
        self._num_hidden_layers = config.num_hidden_layers
        self._num_attention_heads = config.num_key_value_heads if hasattr(config, "num_key_value_heads") else config.num_attention_heads
        self._head_dim = config.head_dim if hasattr(config, "head_dim") else config.hidden_size // config.num_attention_heads
        self._vocab_size = config.vocab_size
        self._cached_len = 0
        self._init_prob_history = torch.empty((INFERENCE_BATCH_SIZE, 0, self._vocab_size)).to(self.device)
        self._prob_history = self._init_prob_history
        self._full_key_values = torch.zeros((self._num_hidden_layers, 2, INFERENCE_BATCH_SIZE, \
                                            self._num_attention_heads, self._max_length, self._head_dim)).to(self.device)
        self._init_key_values = torch.zeros((self._num_hidden_layers, 2, INFERENCE_BATCH_SIZE, \
                                            self._num_attention_heads, self._cached_len, self._head_dim)).to(self.device)
        self._dummy_key_values = self._init_key_values

    @classmethod
    def add_static_cache_to_model(cls, model:AutoModelForCausalLM, cache_config:dict, max_generated_length:int, dtype:torch.dtype, device:torch.device):
        assert "model_config" in cache_config, "you must specify how your model will be updated to accomadate a static kv cache."
        model_config = cache_config["model_config"]
        assert "path_to_blocks" in model_config, "you must specify how to reach the blocks from the model."
        assert "child_ref_in_parent_forward" in model_config, "you must specify how each child module is specified in the forward method of the parent module."
        imports = cache_config["imports"]

        module_with_input_pos = model
        for (prop, forward_prop_ref) in zip(model_config["path_to_blocks"], model_config["child_ref_in_parent_forward"]):
            module_forward_str = get_method_str(module_with_input_pos, "forward")
            module_forward_str_kv_cache = add_input_pos_to_func_str(module_forward_str, forward_prop_ref, "input_pos=input_pos")
            module_forward_str_kv_cache = add_default_parameter(module_forward_str_kv_cache, "forward", "input_pos", "Optional[torch.Tensor]", None, True)
            add_str_as_func(module_with_input_pos, "forward", module_forward_str_kv_cache, imports)

            module_with_input_pos = getattr(module_with_input_pos, prop)
        
        assert isinstance(module_with_input_pos, nn.ModuleList), "Once we finish iterating through 'path_to_blocks', the property that you arrive at must be a nn.ModuleList."
        assert "block_config" in cache_config, "you must specify how your blocks will be updated to accomadate the static kv cache."

        block_config = cache_config["block_config"]
        assert "path_to_attn" in block_config, "you must specify how to reach the attention layer from the blocks."
        assert "child_ref_in_parent_forward" in block_config, "you must specify how each child module is specified in the forward method of the parent module."

        for block in module_with_input_pos:
            sub_module_with_input_pos = block
            for (prop, forward_prop_ref) in zip(block_config["path_to_attn"], block_config["child_ref_in_parent_forward"]):
                sub_module_forward_str = get_method_str(sub_module_with_input_pos, "forward")
                sub_module_forward_str_kv_cache = add_input_pos_to_func_str(sub_module_forward_str, forward_prop_ref, "input_pos=input_pos")
                sub_module_forward_str_kv_cache = add_default_parameter(sub_module_forward_str_kv_cache, "forward", "input_pos", "Optional[torch.Tensor]", None, True)
                add_str_as_func(sub_module_with_input_pos, "forward", sub_module_forward_str_kv_cache, imports)

                sub_module_with_input_pos = getattr(sub_module_with_input_pos, prop)

            assert "attn_config" in cache_config, "you must specify how the attention layer is modified to accomodate the static kv cache."
            attn_config = cache_config["attn_config"]
            assert "cache_update_config" in attn_config, "you must specify how the static kv cache is updated in your attention layer."
            cache_update_config = attn_config["cache_update_config"]

            attention_layer = sub_module_with_input_pos
            cache = KVCache(model.config, 1, max_generated_length, device = device, dtype=dtype)  
            attention_layer.kv_cache = cache

            assert "kv_cache_condition" in cache_update_config, "you must specify the condition under which the key-value cache is updated in the attention layer."
            assert "key_name" in cache_update_config, "you must specify a key name to be updated."
            assert "value_name" in cache_update_config, "you must specify a value name to be updated." 
            key_name, value_name = cache_update_config["key_name"], cache_update_config["value_name"]
            new_key_name, new_value_name = cache_update_config.get("new_key_name", key_name), cache_update_config.get("new_value_name", value_name)
            
            #this will add the kv cache to the attention layer.
            attn_forward_str = get_method_str(attention_layer, "forward")
            attn_forward_str_kv_cache = modify_if_block(attn_forward_str, cache_update_config["kv_cache_condition"], \
                [f'{new_key_name}, {new_value_name} = self.kv_cache.update({key_name}, {value_name}, -1, input_pos=input_pos)'])
            attn_forward_str_kv_cache = add_default_parameter(attn_forward_str_kv_cache, "forward", "input_pos", "Optional[torch.Tensor]", None, True)

            assert "causal_mask_config" in attn_config, "you must specify how the causal mask needs to be updated to accomadate the static kv cache."
            causal_mask_config = attn_config["causal_mask_config"]
            assert "causal_mask_application" in causal_mask_config, "you must specify whether the causal mask is applied conditionally or not."
            assert "causal_mask_method" in causal_mask_config, "you must specify the name of the method where the causal mask is applied to QK^T."

            if causal_mask_config["causal_mask_method"] == "forward":
                attn_forward_str_causal_mask = shift_right(attn_forward_str_kv_cache)
            else:
                add_str_as_func(attention_layer, "forward", attn_forward_str_kv_cache, imports)
                attn_forward_str_causal_mask = get_method_str(attention_layer, causal_mask_config["causal_mask_method"])

            if causal_mask_config["causal_mask_application"] == "conditional":
                assert "causal_mask_condition" in causal_mask_config, "you must specify the condition under which the causal mask is applied to QK^T."

                #this will add the causal mask to the _attn method of the attention layer.
                attn_forward_str_causal_mask_modified = modify_if_block(attn_forward_str_causal_mask, causal_mask_config["causal_mask_condition"], \
                                                        ["causal_mask = self.kv_cache.partial_mask.view(attn_weights.shape)", \
                                                        "attn_weights = attn_weights + causal_mask"])
            else:
                assert "causal_mask_line" in causal_mask_config, "you must specify the first line that you want to overwrite to apply the causal mask to QK^T."
                assert "num_lines" in causal_mask_config, "you must specify how many lines you want to overwrite to apply the causal mask to QK^T."

                attn_forward_str_causal_mask_modified = modify_function_block(attn_forward_str_causal_mask, causal_mask_config["causal_mask_line"], causal_mask_config["num_lines"], \
                                                        ["causal_mask = self.kv_cache.partial_mask.view(attn_weights.shape)", \
                                                        "attn_weights = attn_weights + causal_mask"])
                
            add_str_as_func(attention_layer, causal_mask_config["causal_mask_method"], attn_forward_str_causal_mask_modified, imports)

        return model

    @torch.no_grad()
    def prefill(self, input_ids:torch.Tensor) -> torch.Tensor:
        outputs = self._model(input_ids=input_ids, past_key_values=self._dummy_key_values, input_pos=torch.arange(self._cached_len, input_ids.shape[-1]))
        prob_history = outputs.logits
        self._cached_len = input_ids.shape[-1]
        last_q = prob_history[:, -1, :]
        self._prob_history = torch.cat([self._prob_history, prob_history], dim=-2)
        self._dummy_key_values = self._full_key_values[:, :, :, :, :self._cached_len]
        return last_q

    @torch.no_grad()
    def forward(self, input_ids:torch.Tensor) -> torch.Tensor:
        # return the last token's logits
        input_pos = torch.arange(self._cached_len, self._cached_len + input_ids.shape[-1])
        outputs = self._model(input_ids=input_ids, past_key_values=self._dummy_key_values, input_pos=input_pos)
        prob_history = outputs.logits
        self._cached_len = self._cached_len + input_ids.shape[-1]
        last_q = prob_history[:, -1, :]
        self._prob_history = torch.cat([self._prob_history, prob_history], dim=-2)
        self._dummy_key_values = self._full_key_values[:, :, :, :, :self._cached_len]
        return last_q

    @torch.no_grad()
    def generate(self, prefix:torch.Tensor, gamma:int, **sampling_kwargs) -> torch.Tensor:
        """ forward the model gamma times

        Args:
            prefix (torch.Tensor): the prefix
            gamma (int): how many times approx guesses

        Returns:
            Torch.Tensor: prefix+generated tokens
        """
        x = prefix

        #prefill stage
        last_logits = self.prefill(x)
        last_prob = torch.nn.functional.softmax(last_logits, dim=-1)
        next_tok = self.sample(last_prob, **sampling_kwargs)
        x = torch.cat((x, next_tok), dim=1)

        for _ in range(gamma):
            last_logits = self.forward(next_tok)
            last_prob = torch.nn.functional.softmax(last_logits, dim=-1)
            next_tok = self.sample(last_prob, **sampling_kwargs)
            x = torch.cat((x, next_tok), dim=1)
        
        self.clear()

        return x

    @torch.no_grad()
    def decode_function(self, uncached_input_ids:torch.Tensor, length:int, return_text:bool=True, **sampling_kwargs) -> torch.Tensor:

        all_input_ids = torch.empty((INFERENCE_BATCH_SIZE, 0), dtype=torch.long).to(self.device)

        for _ in range(length):
            with torch.no_grad():
                last_logits = self.forward(uncached_input_ids)

            token_probabilities = torch.nn.functional.softmax(last_logits, dim=-1)
            next_token = self.sample(token_probabilities, **sampling_kwargs)
            # Append the sampled token to the input sequence
            uncached_input_ids = next_token
            all_input_ids = torch.cat([all_input_ids, next_token], dim=-1)
        
        all_probabilities = torch.nn.functional.softmax(self._prob_history[:, -length-1:-1, :], dim=-1).squeeze(0)

        if return_text:
            return all_input_ids, all_probabilities
        else:
            return all_probabilities
        
    @torch.no_grad()
    def rollback_cache(self, last_pos:int) -> None:
        self._cached_len = last_pos
        self._dummy_key_values = self._full_key_values[:, :, :, :, :self._cached_len]
        self._prob_history = self._prob_history[:, :self._cached_len, :]
    
    @torch.no_grad()
    def clear(self):
        self._dummy_key_values = self._init_key_values
        self._prob_history = self._init_prob_history
        self._cached_len = 0

def add_kv_cache(transformer:nn.Module, sampling_fn:Callable, max_length:int, cache_config:dict, dtype) -> KVCacheModel:
    model = KVCacheModel(transformer, sampling_fn, max_length, cache_config, dtype)
    return model


from .shift_left import shift_left

def modify_if_block(method_str:str, condition:str, new_condition_strings_list: list[str]):

    # Find the index of the if condition
    if_condition_index = method_str.find(f"{condition}:")

    if if_condition_index != -1:
        # Extract the line containing the if condition
        lines = method_str.split('\n')
        for line_num, line in enumerate(lines):
            if condition in line:
                if_condition_line = line
                left_index = line_num
                break

        # Find the indentation level of the if condition
        if_condition_indentation = len(if_condition_line) - len(if_condition_line.lstrip())

        # Find the index where the indentation level decreases
        for right_index, line in enumerate(lines[left_index+1:], start=left_index+1):
            if line.strip() and len(line) - len(line.lstrip()) <= if_condition_indentation:
                break

        # Calculate the indentation on lines[right_index]
        if_condition_indentation = len(lines[right_index]) - len(lines[right_index].lstrip())

        # Shift new_condition_strings_list to the right by if_condition_indentation + 4
        total_spaces = (if_condition_indentation + 4) * " "
        for i in range(len(new_condition_strings_list)):
            new_condition_strings_list[i] = total_spaces + new_condition_strings_list[i]

        # Concatenate the modified block
        lines = lines[:left_index + 1] + new_condition_strings_list + lines[right_index:]
        method_str = "\n".join(lines)

    method_str = shift_left(method_str)
    return method_str


from setuptools import setup, find_packages

setup(
    name='gptfast',
    version='0.2.0',
    author="MDK8888",
    description="Accelerate transformer inference by 6-8.5x. Native to Huggingface and PyTorch.",
    packages=find_packages(),
    install_requires=['torch==2.2.1', 'sympy==1.12', 'typing-extensions==4.9.0', 'networkx==3.2.1', 'jinja2==3.1.3', 'triton==2.2.0', 'fsspec==2023.10.0', 'filelock==3.13.1', 'MarkupSafe==2.1.4', 'mpmath==1.3.0', 'transformers==4.39.2', 'tqdm==4.66.1', 'pyyaml==5.3.1', 'safetensors==0.4.1', 'numpy==1.26.3', 'huggingface-hub==0.20.2', 'tokenizers==0.15.0', 'packaging==23.2', 'regex==2023.12.25', 'requests==2.31.0', 'bitsandbytes==0.43.0', 'accelerate==0.27.2'],
    license="Apache License 2.0",
    package_data={
        '':["LICENSE", "requirements.txt"]
    }
)


from typing import Callable
import torch
from transformers import AutoModelForCausalLM
from GPTFast.Core.KVCache import add_kv_cache
from GPTFast.Core.Compile import torch_compile_model
from GPTFast.Core.Decode import add_speculative_decoding
from GPTFast.Core.Quantize import load_int8

def gpt_fast(model_name:str, sample_function:Callable, max_length:int, cache_config:dict, **spec_dec_kwargs):
    model = load_int8(model_name)
    model = add_kv_cache(model, sample_function, max_length, cache_config, dtype=torch.float16)
    spec_decode = False
    if spec_dec_kwargs:
        draft_model_name = spec_dec_kwargs.pop("draft_model_name")
        draft_model = AutoModelForCausalLM.from_pretrained(draft_model_name)
        draft_model = add_kv_cache(draft_model, sample_function, max_length, cache_config, dtype=torch.float32)
        model = add_speculative_decoding(model, draft_model, **spec_dec_kwargs)
        spec_decode = True
    model = torch_compile_model(model, spec_decode=spec_decode)
    return model


import inspect

def get_method_str(object_instance:object, method_name:str) -> str:
    if not hasattr(object_instance, method_name):
        return f"Method '{method_name}' not found in the object."

    # Get the original method
    original_method = getattr(object_instance, method_name)

    # Convert the method to a string
    method_str = inspect.getsource(original_method)

    return method_str


import torch
import torch.nn as nn

def torch_compile_model(model:nn.Module, spec_decode:bool = False) -> nn.Module:
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = torch.compile(model, mode="reduce-overhead", fullgraph=True)
    #model.prefill = types.MethodType(torch.compile(model.prefill, dynamic=True, fullgraph=True), model)
    model = model.to(device)
    if spec_decode:
        assert hasattr(model, "draft_model"), "You have passed spec_decode = True in your torch_compile but your model doesn't have a draft model."
        draft_model = model.draft_model
        draft_model = torch.compile(draft_model, mode="reduce-overhead", fullgraph=True)
        #draft_model.prefill = types.MethodType(torch.compile(draft_model.prefill, dynamic=True, fullgraph=True), draft_model)
        draft_model = draft_model.to(device)
    return model